from ..constants import BusType
from ..errors import InvalidAddressError

from urllib.parse import unquote
import re
import os

invalid_address_chars_re = re.compile(r'[^-0-9A-Za-z_/.%]')


def parse_address(address_str):
    addresses = []

    for address in filter(lambda a: a, address_str.split(';')):
        if address.find(':') == -1:
            raise InvalidAddressError('address did not contain a transport')

        transport, opt_string = address.split(':', 1)
        options = {}

        for kv in filter(lambda s: s, opt_string.split(',')):
            if kv.find('=') == -1:
                raise InvalidAddressError('address option did not contain a value')
            k, v = kv.split('=', 1)
            if invalid_address_chars_re.search(v):
                raise InvalidAddressError('address contains invalid characters')
            # XXX the actual unquote rules are simpler than this
            v = unquote(v)
            options[k] = v

        addresses.append((transport, options))

    if not addresses:
        raise InvalidAddressError(f'address string contained no addresses: "{address_str}"')

    return addresses


def get_system_bus_address():
    if 'DBUS_SYSTEM_BUS_ADDRESS' in os.environ:
        return os.environ['DBUS_SYSTEM_BUS_ADDRESS']
    else:
        return 'unix:path=/var/run/dbus/system_bus_socket'


display_re = re.compile(r'.*:([0-9]+)\.?.*')
remove_quotes_re = re.compile(r'''^['"]?(.*?)['"]?$''')


def get_session_bus_address():
    if 'DBUS_SESSION_BUS_ADDRESS' in os.environ:
        return os.environ['DBUS_SESSION_BUS_ADDRESS']

    home = os.environ['HOME']
    if 'DISPLAY' not in os.environ:
        raise InvalidAddressError(
            'DBUS_SESSION_BUS_ADDRESS not set and could not get DISPLAY environment variable to get bus address'
        )

    display = os.environ['DISPLAY']
    try:
        display = display_re.search(display).group(1)
    except Exception:
        raise InvalidAddressError(
            f'DBUS_SESSION_BUS_ADDRESS not set and could not parse DISPLAY environment variable to get bus address: {display}'
        )

    # XXX: this will block but they're very small files and fs operations
    # should be fairly reliable. fix this by passing in an async func to read
    # the file for each io backend.
    machine_id = None
    with open('/var/lib/dbus/machine-id') as f:
        machine_id = f.read().rstrip()

    dbus_info_file_name = f'{home}/.dbus/session-bus/{machine_id}-{display}'
    dbus_info = None
    try:
        with open(dbus_info_file_name) as f:
            dbus_info = f.read().rstrip()
    except Exception:
        raise InvalidAddressError(f'could not open dbus info file: {dbus_info_file_name}')

    for line in dbus_info.split('\n'):
        if line.strip().startswith('DBUS_SESSION_BUS_ADDRESS='):
            _, addr = line.split('=', 1)
            if not addr:
                raise InvalidAddressError(
                    f'DBUS_SESSION_BUS_ADDRESS variable not set correctly in dbus info file: {dbus_info_file_name}'
                )
            addr = remove_quotes_re.search(addr).group(1)
            return addr

    raise InvalidAddressError('could not find dbus session bus address')


def get_bus_address(bus_type):
    if bus_type == BusType.SESSION:
        return get_session_bus_address()
    elif bus_type == BusType.SYSTEM:
        return get_system_bus_address()
    else:
        raise Exception('got unknown bus type: {bus_type}')



from typing import List, Any, Union
import inspect
from ..signature import SignatureTree, Variant
import ast


def signature_contains_type(signature: Union[str, SignatureTree], body: List[Any],
                            token: str) -> bool:
    '''For a given signature and body, check to see if it contains any members
    with the given token'''
    if type(signature) is str:
        signature = SignatureTree._get(signature)

    queue = []
    contains_variants = False
    for st in signature.types:
        queue.append(st)

    while True:
        if not queue:
            break
        st = queue.pop()
        if st.token == token:
            return True
        elif st.token == 'v':
            contains_variants = True
        queue.extend(st.children)

    if not contains_variants:
        return False

    for member in body:
        queue.append(member)

    while True:
        if not queue:
            return False
        member = queue.pop()
        if type(member) is Variant and \
                signature_contains_type(member.signature, [member.value], token):
            return True
        elif type(member) is list:
            queue.extend(member)
        elif type(member) is dict:
            queue.extend(member.values())


def replace_fds_with_idx(signature: Union[str, SignatureTree],
                         body: List[Any]) -> (List[Any], List[int]):
    '''Take the high level body format and convert it into the low level body
    format. Type 'h' refers directly to the fd in the body. Replace that with
    an index and return the corresponding list of unix fds that can be set on
    the Message'''
    if type(signature) is str:
        signature = SignatureTree._get(signature)

    if not signature_contains_type(signature, body, 'h'):
        return body, []

    unix_fds = []

    def _replace(fd):
        try:
            return unix_fds.index(fd)
        except ValueError:
            unix_fds.append(fd)
            return len(unix_fds) - 1

    _replace_fds(body, signature.types, _replace)

    return body, unix_fds


def replace_idx_with_fds(signature: Union[str, SignatureTree], body: List[Any],
                         unix_fds: List[int]) -> List[Any]:
    '''Take the low level body format and return the high level body format.
    Type 'h' refers to an index in the unix_fds array. Replace those with the
    actual file descriptor or `None` if one does not exist.'''
    if type(signature) is str:
        signature = SignatureTree._get(signature)

    if not signature_contains_type(signature, body, 'h'):
        return body

    def _replace(idx):
        try:
            return unix_fds[idx]
        except IndexError:
            return None

    _replace_fds(body, signature.types, _replace)

    return body


def parse_annotation(annotation: str) -> str:
    '''
    Because of PEP 563, if `from __future__ import annotations` is used in code
    or on Python version >=3.10 where this is the default, return annotations
    from the `inspect` module will return annotations as "forward definitions".
    In this case, we must eval the result which we do only when given a string
    constant.
    '''
    def raise_value_error():
        raise ValueError(f'service annotations must be a string constant (got {annotation})')

    if not annotation or annotation is inspect.Signature.empty:
        return ''
    if type(annotation) is not str:
        raise_value_error()
    try:
        body = ast.parse(annotation).body
        if len(body) == 1 and type(body[0].value) is ast.Constant:
            if type(body[0].value.value) is not str:
                raise_value_error()
            return body[0].value.value
    except SyntaxError:
        pass

    return annotation


def _replace_fds(body_obj: List[Any], children, replace_fn):
    '''Replace any type 'h' with the value returned by replace_fn() given the
    value of the fd field. This is used by the high level interfaces which
    allow type 'h' to be the fd directly instead of an index in an external
    array such as in the spec.'''
    for index, st in enumerate(children):
        if not any(sig in st.signature for sig in 'hv'):
            continue
        if st.signature == 'h':
            body_obj[index] = replace_fn(body_obj[index])
        elif st.token == 'a':
            if st.children[0].token == '{':
                _replace_fds(body_obj[index], st.children, replace_fn)
            else:
                for i, child in enumerate(body_obj[index]):
                    if st.signature == 'ah':
                        body_obj[index][i] = replace_fn(child)
                    else:
                        _replace_fds([child], st.children, replace_fn)
        elif st.token in '(':
            _replace_fds(body_obj[index], st.children, replace_fn)
        elif st.token in '{':
            for key, value in list(body_obj.items()):
                body_obj.pop(key)
                if st.children[0].signature == 'h':
                    key = replace_fn(key)
                if st.children[1].signature == 'h':
                    value = replace_fn(value)
                else:
                    _replace_fds([value], [st.children[1]], replace_fn)
                body_obj[key] = value

        elif st.signature == 'v':
            if body_obj[index].signature == 'h':
                body_obj[index].value = replace_fn(body_obj[index].value)
            else:
                _replace_fds([body_obj[index].value], [body_obj[index].type], replace_fn)

        elif st.children:
            _replace_fds(body_obj[index], st.children, replace_fn)



from .constants import PropertyAccess
from .signature import SignatureTree, SignatureBodyMismatchError, Variant
from . import introspection as intr
from .errors import SignalDisabledError
from ._private.util import signature_contains_type, replace_fds_with_idx, replace_idx_with_fds, parse_annotation

from functools import wraps
import inspect
from typing import no_type_check_decorator, Dict, List, Any
import copy
import asyncio


class _Method:
    def __init__(self, fn, name, disabled=False):
        in_signature = ''
        out_signature = ''

        inspection = inspect.signature(fn)

        in_args = []
        for i, param in enumerate(inspection.parameters.values()):
            if i == 0:
                # first is self
                continue
            annotation = parse_annotation(param.annotation)
            if not annotation:
                raise ValueError(
                    'method parameters must specify the dbus type string as an annotation')
            in_args.append(intr.Arg(annotation, intr.ArgDirection.IN, param.name))
            in_signature += annotation

        out_args = []
        out_signature = parse_annotation(inspection.return_annotation)
        if out_signature:
            for type_ in SignatureTree._get(out_signature).types:
                out_args.append(intr.Arg(type_, intr.ArgDirection.OUT))

        self.name = name
        self.fn = fn
        self.disabled = disabled
        self.introspection = intr.Method(name, in_args, out_args)
        self.in_signature = in_signature
        self.out_signature = out_signature
        self.in_signature_tree = SignatureTree._get(in_signature)
        self.out_signature_tree = SignatureTree._get(out_signature)


def method(name: str = None, disabled: bool = False):
    """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus service method.

    The parameters and return value must each be annotated with a signature
    string of a single complete DBus type.

    This class method will be called when a client calls the method on the DBus
    interface. The parameters given to the function come from the calling
    client and will conform to the dbus-next type system. The parameters
    returned will be returned to the calling client and must conform to the
    dbus-next type system. If multiple parameters are returned, they must be
    contained within a :class:`list`.

    The decorated method may raise a :class:`DBusError <dbus_next.DBusError>`
    to return an error to the client.

    :param name: The member name that DBus clients will use to call this method. Defaults to the name of the class method.
    :type name: str
    :param disabled: If set to true, the method will not be visible to clients.
    :type disabled: bool

    :example:

    ::

        @method()
        def echo(self, val: 's') -> 's':
            return val

        @method()
        def echo_two(self, val1: 's', val2: 'u') -> 'su':
            return [val1, val2]
    """
    if name is not None and type(name) is not str:
        raise TypeError('name must be a string')
    if type(disabled) is not bool:
        raise TypeError('disabled must be a bool')

    @no_type_check_decorator
    def decorator(fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            fn(*args, **kwargs)

        fn_name = name if name else fn.__name__
        wrapped.__dict__['__DBUS_METHOD'] = _Method(fn, fn_name, disabled=disabled)

        return wrapped

    return decorator


class _Signal:
    def __init__(self, fn, name, disabled=False):
        inspection = inspect.signature(fn)

        args = []
        signature = ''
        signature_tree = None

        return_annotation = parse_annotation(inspection.return_annotation)

        if return_annotation:
            signature = return_annotation
            signature_tree = SignatureTree._get(signature)
            for type_ in signature_tree.types:
                args.append(intr.Arg(type_, intr.ArgDirection.OUT))
        else:
            signature = ''
            signature_tree = SignatureTree._get('')

        self.signature = signature
        self.signature_tree = signature_tree
        self.name = name
        self.disabled = disabled
        self.introspection = intr.Signal(self.name, args)


def signal(name: str = None, disabled: bool = False):
    """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus signal.

    The signal is broadcast on the bus when the decorated class method is
    called by the user.

    If the signal has an out argument, the class method must have a return type
    annotation with a signature string of a single complete DBus type and the
    return value of the class method must conform to the dbus-next type system.
    If the signal has multiple out arguments, they must be returned within a
    ``list``.

    :param name: The member name that will be used for this signal. Defaults to
        the name of the class method.
    :type name: str
    :param disabled: If set to true, the signal will not be visible to clients.
    :type disabled: bool

    :example:

    ::

        @signal()
        def string_signal(self, val) -> 's':
            return val

        @signal()
        def two_strings_signal(self, val1, val2) -> 'ss':
            return [val1, val2]
    """
    if name is not None and type(name) is not str:
        raise TypeError('name must be a string')
    if type(disabled) is not bool:
        raise TypeError('disabled must be a bool')

    @no_type_check_decorator
    def decorator(fn):
        fn_name = name if name else fn.__name__
        signal = _Signal(fn, fn_name, disabled)

        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            if signal.disabled:
                raise SignalDisabledError('Tried to call a disabled signal')
            result = fn(self, *args, **kwargs)
            ServiceInterface._handle_signal(self, signal, result)
            return result

        wrapped.__dict__['__DBUS_SIGNAL'] = signal

        return wrapped

    return decorator


class _Property(property):
    def set_options(self, options):
        self.options = getattr(self, 'options', {})
        for k, v in options.items():
            self.options[k] = v

        if 'name' in options and options['name'] is not None:
            self.name = options['name']
        else:
            self.name = self.prop_getter.__name__

        if 'access' in options:
            self.access = PropertyAccess(options['access'])
        else:
            self.access = PropertyAccess.READWRITE

        if 'disabled' in options:
            self.disabled = options['disabled']
        else:
            self.disabled = False

        self.introspection = intr.Property(self.name, self.signature, self.access)

        self.__dict__['__DBUS_PROPERTY'] = True

    def __init__(self, fn, *args, **kwargs):
        self.prop_getter = fn
        self.prop_setter = None

        inspection = inspect.signature(fn)
        if len(inspection.parameters) != 1:
            raise ValueError('the property must only have the "self" input parameter')

        return_annotation = parse_annotation(inspection.return_annotation)

        if not return_annotation:
            raise ValueError(
                'the property must specify the dbus type string as a return annotation string')

        self.signature = return_annotation
        tree = SignatureTree._get(return_annotation)

        if len(tree.types) != 1:
            raise ValueError('the property signature must be a single complete type')

        self.type = tree.types[0]

        if 'options' in kwargs:
            options = kwargs['options']
            self.set_options(options)
            del kwargs['options']

        super().__init__(fn, *args, **kwargs)

    def setter(self, fn, **kwargs):
        # XXX The setter decorator seems to be recreating the class in the list
        # of class members and clobbering the options so we need to reset them.
        # Why does it do that?
        result = super().setter(fn, **kwargs)
        result.prop_setter = fn
        result.set_options(self.options)
        return result


def dbus_property(access: PropertyAccess = PropertyAccess.READWRITE,
                  name: str = None,
                  disabled: bool = False):
    """A decorator to mark a class method of a :class:`ServiceInterface` to be a DBus property.

    The class method must be a Python getter method with a return annotation
    that is a signature string of a single complete DBus type. When a client
    gets the property through the ``org.freedesktop.DBus.Properties``
    interface, the getter will be called and the resulting value will be
    returned to the client.

    If the property is writable, it must have a setter method that takes a
    single parameter that is annotated with the same signature. When a client
    sets the property through the ``org.freedesktop.DBus.Properties``
    interface, the setter will be called with the value from the calling
    client.

    The parameters of the getter and the setter must conform to the dbus-next
    type system. The getter or the setter may raise a :class:`DBusError
    <dbus_next.DBusError>` to return an error to the client.

    :param name: The name that DBus clients will use to interact with this
        property on the bus.
    :type name: str
    :param disabled: If set to true, the property will not be visible to
        clients.
    :type disabled: bool

    :example:

    ::

        @dbus_property()
        def string_prop(self) -> 's':
            return self._string_prop

        @string_prop.setter
        def string_prop(self, val: 's'):
            self._string_prop = val
    """
    if type(access) is not PropertyAccess:
        raise TypeError('access must be a PropertyAccess class')
    if name is not None and type(name) is not str:
        raise TypeError('name must be a string')
    if type(disabled) is not bool:
        raise TypeError('disabled must be a bool')

    @no_type_check_decorator
    def decorator(fn):
        options = {'name': name, 'access': access, 'disabled': disabled}
        return _Property(fn, options=options)

    return decorator


class ServiceInterface:
    """An abstract class that can be extended by the user to define DBus services.

    Instances of :class:`ServiceInterface` can be exported on a path of the bus
    with the :class:`export <dbus_next.message_bus.BaseMessageBus.export>`
    method of a :class:`MessageBus <dbus_next.message_bus.BaseMessageBus>`.

    Use the :func:`@method <dbus_next.service.method>`, :func:`@dbus_property
    <dbus_next.service.dbus_property>`, and :func:`@signal
    <dbus_next.service.signal>` decorators to mark class methods as DBus
    methods, properties, and signals respectively.

    :ivar name: The name of this interface as it appears to clients. Must be a
        valid interface name.
    :vartype name: str
    """
    def __init__(self, name: str):
        # TODO cannot be overridden by a dbus member
        self.name = name
        self.__methods = []
        self.__properties = []
        self.__signals = []
        self.__buses = set()

        for name, member in inspect.getmembers(type(self)):
            member_dict = getattr(member, '__dict__', {})
            if type(member) is _Property:
                # XXX The getter and the setter may show up as different
                # members if they have different names. But if they have the
                # same name, they will be the same member. So we try to merge
                # them together here. I wish we could make this cleaner.
                found = False
                for prop in self.__properties:
                    if prop.prop_getter is member.prop_getter:
                        found = True
                        if member.prop_setter is not None:
                            prop.prop_setter = member.prop_setter

                if not found:
                    self.__properties.append(member)
            elif '__DBUS_METHOD' in member_dict:
                method = member_dict['__DBUS_METHOD']
                assert type(method) is _Method
                self.__methods.append(method)
            elif '__DBUS_SIGNAL' in member_dict:
                signal = member_dict['__DBUS_SIGNAL']
                assert type(signal) is _Signal
                self.__signals.append(signal)

        # validate that writable properties have a setter
        for prop in self.__properties:
            if prop.access.writable() and prop.prop_setter is None:
                raise ValueError(f'property "{prop.name}" is writable but does not have a setter')

    def emit_properties_changed(self,
                                changed_properties: Dict[str, Any],
                                invalidated_properties: List[str] = []):
        """Emit the ``org.freedesktop.DBus.Properties.PropertiesChanged`` signal.

        This signal is intended to be used to alert clients when a property of
        the interface has changed.

        :param changed_properties: The keys must be the names of properties exposed by this bus. The values must be valid for the signature of those properties.
        :type changed_properties: dict(str, Any)
        :param invalidated_properties: A list of names of properties that are now invalid (presumably for clients who cache the value).
        :type invalidated_properties: list(str)
        """
        # TODO cannot be overridden by a dbus member
        variant_dict = {}

        for prop in ServiceInterface._get_properties(self):
            if prop.name in changed_properties:
                variant_dict[prop.name] = Variant(prop.signature, changed_properties[prop.name])

        body = [self.name, variant_dict, invalidated_properties]
        for bus in ServiceInterface._get_buses(self):
            bus._interface_signal_notify(self, 'org.freedesktop.DBus.Properties',
                                         'PropertiesChanged', 'sa{sv}as', body)

    def introspect(self) -> intr.Interface:
        """Get introspection information for this interface.

        This might be useful for creating clients for the interface or examining the introspection output of an interface.

        :returns: The introspection data for the interface.
        :rtype: :class:`dbus_next.introspection.Interface`
        """
        # TODO cannot be overridden by a dbus member
        return intr.Interface(self.name,
                              methods=[
                                  method.introspection
                                  for method in ServiceInterface._get_methods(self)
                                  if not method.disabled
                              ],
                              signals=[
                                  signal.introspection
                                  for signal in ServiceInterface._get_signals(self)
                                  if not signal.disabled
                              ],
                              properties=[
                                  prop.introspection
                                  for prop in ServiceInterface._get_properties(self)
                                  if not prop.disabled
                              ])

    @staticmethod
    def _get_properties(interface):
        return interface.__properties

    @staticmethod
    def _get_methods(interface):
        return interface.__methods

    @staticmethod
    def _get_signals(interface):
        return interface.__signals

    @staticmethod
    def _get_buses(interface):
        return interface.__buses

    @staticmethod
    def _add_bus(interface, bus):
        interface.__buses.add(bus)

    @staticmethod
    def _remove_bus(interface, bus):
        interface.__buses.remove(bus)

    @staticmethod
    def _msg_body_to_args(msg):
        if signature_contains_type(msg.signature_tree, msg.body, 'h'):
            # XXX: This deep copy could be expensive if messages are very
            # large. We could optimize this by only copying what we change
            # here.
            return replace_idx_with_fds(msg.signature_tree, copy.deepcopy(msg.body), msg.unix_fds)
        else:
            return msg.body

    @staticmethod
    def _fn_result_to_body(result, signature_tree):
        '''The high level interfaces may return single values which may be
        wrapped in a list to be a message body. Also they may return fds
        directly for type 'h' which need to be put into an external list.'''
        out_len = len(signature_tree.types)
        if result is None:
            result = []
        else:
            if out_len == 1:
                result = [result]
            else:
                if type(result) is not list:
                    raise SignatureBodyMismatchError(
                        'Expected signal to return a list of arguments')

        if out_len != len(result):
            raise SignatureBodyMismatchError(
                f"Signature and function return mismatch, expected {len(signature_tree.types)} arguments but got {len(result)}"
            )

        return replace_fds_with_idx(signature_tree, result)

    @staticmethod
    def _handle_signal(interface, signal, result):
        body, fds = ServiceInterface._fn_result_to_body(result, signal.signature_tree)
        for bus in ServiceInterface._get_buses(interface):
            bus._interface_signal_notify(interface, interface.name, signal.name, signal.signature,
                                         body, fds)

    @staticmethod
    def _get_property_value(interface, prop, callback):
        # XXX MUST CHECK TYPE RETURNED BY GETTER
        try:
            if asyncio.iscoroutinefunction(prop.prop_getter):
                task = asyncio.ensure_future(prop.prop_getter(interface))

                def get_property_callback(task):
                    try:
                        result = task.result()
                    except Exception as e:
                        callback(interface, prop, None, e)
                        return

                    callback(interface, prop, result, None)

                task.add_done_callback(get_property_callback)
                return

            callback(interface, prop, getattr(interface, prop.prop_getter.__name__), None)
        except Exception as e:
            callback(interface, prop, None, e)

    @staticmethod
    def _set_property_value(interface, prop, value, callback):
        # XXX MUST CHECK TYPE TO SET
        try:
            if asyncio.iscoroutinefunction(prop.prop_setter):
                task = asyncio.ensure_future(prop.prop_setter(interface, value))

                def set_property_callback(task):
                    try:
                        task.result()
                    except Exception as e:
                        callback(interface, prop, e)
                        return

                    callback(interface, prop, None)

                task.add_done_callback(set_property_callback)
                return

            setattr(interface, prop.prop_setter.__name__, value)
            callback(interface, prop, None)
        except Exception as e:
            callback(interface, prop, e)

    @staticmethod
    def _get_all_property_values(interface, callback, user_data=None):
        result = {}
        result_error = None

        for prop in ServiceInterface._get_properties(interface):
            if prop.disabled or not prop.access.readable():
                continue
            result[prop.name] = None

        if not result:
            callback(interface, result, user_data, None)
            return

        def get_property_callback(interface, prop, value, e):
            nonlocal result_error
            if e is not None:
                result_error = e
                del result[prop.name]
            else:
                try:
                    result[prop.name] = Variant(prop.signature, value)
                except SignatureBodyMismatchError as e:
                    result_error = e
                    del result[prop.name]

            if any(v is None for v in result.values()):
                return

            callback(interface, result, user_data, result_error)

        for prop in ServiceInterface._get_properties(interface):
            if prop.disabled or not prop.access.readable():
                continue
            ServiceInterface._get_property_value(interface, prop, get_property_callback)



import re
from .errors import InvalidBusNameError, InvalidObjectPathError, InvalidInterfaceNameError, InvalidMemberNameError

_bus_name_re = re.compile(r'^[A-Za-z_-][A-Za-z0-9_-]*$')
_path_re = re.compile(r'^[A-Za-z0-9_]+$')
_element_re = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')


def is_bus_name_valid(name: str) -> bool:
    """Whether this is a valid bus name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-bus

    :param name: The bus name to validate.
    :type name: str

    :returns: Whether the name is a valid bus name.
    :rtype: bool
    """
    if not isinstance(name, str):
        return False

    if not name or len(name) > 255:
        return False

    if name.startswith(':'):
        # a unique bus name
        return True

    if name.startswith('.'):
        return False

    if name.find('.') == -1:
        return False

    for element in name.split('.'):
        if _bus_name_re.search(element) is None:
            return False

    return True


def is_object_path_valid(path: str) -> bool:
    """Whether this is a valid object path.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-marshaling-object-path

    :param path: The object path to validate.
    :type path: str

    :returns: Whether the object path is valid.
    :rtype: bool
    """
    if not isinstance(path, str):
        return False

    if not path:
        return False

    if not path.startswith('/'):
        return False

    if len(path) == 1:
        return True

    for element in path[1:].split('/'):
        if _path_re.search(element) is None:
            return False

    return True


def is_interface_name_valid(name: str) -> bool:
    """Whether this is a valid interface name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-interface

    :param name: The interface name to validate.
    :type name: str

    :returns: Whether the name is a valid interface name.
    :rtype: bool
    """
    if not isinstance(name, str):
        return False

    if not name or len(name) > 255:
        return False

    if name.startswith('.'):
        return False

    if name.find('.') == -1:
        return False

    for element in name.split('.'):
        if _element_re.search(element) is None:
            return False

    return True


def is_member_name_valid(member: str) -> bool:
    """Whether this is a valid member name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-member

    :param member: The member name to validate.
    :type member: str

    :returns: Whether the name is a valid member name.
    :rtype: bool
    """
    if not isinstance(member, str):
        return False

    if not member or len(member) > 255:
        return False

    if _element_re.search(member) is None:
        return False

    return True


def assert_bus_name_valid(name: str):
    """Raise an error if this is not a valid bus name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-bus

    :param name: The bus name to validate.
    :type name: str

    :raises:
        - :class:`InvalidBusNameError` - If this is not a valid bus name.
    """
    if not is_bus_name_valid(name):
        raise InvalidBusNameError(name)


def assert_object_path_valid(path: str):
    """Raise an error if this is not a valid object path.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-marshaling-object-path

    :param path: The object path to validate.
    :type path: str

    :raises:
        - :class:`InvalidObjectPathError` - If this is not a valid object path.
    """
    if not is_object_path_valid(path):
        raise InvalidObjectPathError(path)


def assert_interface_name_valid(name: str):
    """Raise an error if this is not a valid interface name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-interface

    :param name: The interface name to validate.
    :type name: str

    :raises:
        - :class:`InvalidInterfaceNameError` - If this is not a valid object path.
    """
    if not is_interface_name_valid(name):
        raise InvalidInterfaceNameError(name)


def assert_member_name_valid(member):
    """Raise an error if this is not a valid member name.

    .. seealso:: https://dbus.freedesktop.org/doc/dbus-specification.html#message-protocol-names-member

    :param member: The member name to validate.
    :type member: str

    :raises:
        - :class:`InvalidMemberNameError` - If this is not a valid object path.
    """
    if not is_member_name_valid(member):
        raise InvalidMemberNameError(member)



class SignatureBodyMismatchError(ValueError):
    pass


class InvalidSignatureError(ValueError):
    pass


class InvalidAddressError(ValueError):
    pass


class AuthError(Exception):
    pass


class InvalidMessageError(ValueError):
    pass


class InvalidIntrospectionError(ValueError):
    pass


class InterfaceNotFoundError(Exception):
    pass


class SignalDisabledError(Exception):
    pass


class InvalidBusNameError(TypeError):
    def __init__(self, name):
        super().__init__(f'invalid bus name: {name}')


class InvalidObjectPathError(TypeError):
    def __init__(self, path):
        super().__init__(f'invalid object path: {path}')


class InvalidInterfaceNameError(TypeError):
    def __init__(self, name):
        super().__init__(f'invalid interface name: {name}')


class InvalidMemberNameError(TypeError):
    def __init__(self, member):
        super().__init__(f'invalid member name: {member}')


from .message import Message
from .validators import assert_interface_name_valid
from .constants import ErrorType, MessageType


class DBusError(Exception):
    def __init__(self, type_, text, reply=None):
        super().__init__(text)

        if type(type_) is ErrorType:
            type_ = type_.value

        assert_interface_name_valid(type_)
        if reply is not None and type(reply) is not Message:
            raise TypeError('reply must be of type Message')

        self.type = type_
        self.text = text
        self.reply = reply

    @staticmethod
    def _from_message(msg):
        assert msg.message_type == MessageType.ERROR
        return DBusError(msg.error_name, msg.body[0], reply=msg)

    def _as_message(self, msg):
        return Message.new_error(msg, self.type, self.text)



from ..signature import SignatureTree
from struct import pack


class Marshaller:
    def __init__(self, signature, body):
        self.signature_tree = SignatureTree._get(signature)
        self.signature_tree.verify(body)
        self.buffer = bytearray()
        self.body = body

        self.writers = {
            'y': self.write_byte,
            'b': self.write_boolean,
            'n': self.write_int16,
            'q': self.write_uint16,
            'i': self.write_int32,
            'u': self.write_uint32,
            'x': self.write_int64,
            't': self.write_uint64,
            'd': self.write_double,
            'h': self.write_uint32,
            'o': self.write_string,
            's': self.write_string,
            'g': self.write_signature,
            'a': self.write_array,
            '(': self.write_struct,
            '{': self.write_dict_entry,
            'v': self.write_variant
        }

    def align(self, n):
        offset = n - len(self.buffer) % n
        if offset == 0 or offset == n:
            return 0
        self.buffer.extend(bytes(offset))
        return offset

    def write_byte(self, byte, _=None):
        self.buffer.append(byte)
        return 1

    def write_boolean(self, boolean, _=None):
        if boolean:
            return self.write_uint32(1)
        else:
            return self.write_uint32(0)

    def write_int16(self, int16, _=None):
        written = self.align(2)
        self.buffer.extend(pack('<h', int16))
        return written + 2

    def write_uint16(self, uint16, _=None):
        written = self.align(2)
        self.buffer.extend(pack('<H', uint16))
        return written + 2

    def write_int32(self, int32, _):
        written = self.align(4)
        self.buffer.extend(pack('<i', int32))
        return written + 4

    def write_uint32(self, uint32, _=None):
        written = self.align(4)
        self.buffer.extend(pack('<I', uint32))
        return written + 4

    def write_int64(self, int64, _=None):
        written = self.align(8)
        self.buffer.extend(pack('<q', int64))
        return written + 8

    def write_uint64(self, uint64, _=None):
        written = self.align(8)
        self.buffer.extend(pack('<Q', uint64))
        return written + 8

    def write_double(self, double, _=None):
        written = self.align(8)
        self.buffer.extend(pack('<d', double))
        return written + 8

    def write_signature(self, signature, _=None):
        signature = signature.encode()
        signature_len = len(signature)
        self.buffer.append(signature_len)
        self.buffer.extend(signature)
        self.buffer.append(0)
        return signature_len + 2

    def write_string(self, value, _=None):
        value = value.encode()
        value_len = len(value)
        written = self.write_uint32(value_len)
        self.buffer.extend(value)
        written += value_len
        self.buffer.append(0)
        written += 1
        return written

    def write_variant(self, variant, _=None):
        written = self.write_signature(variant.signature)
        written += self.write_single(variant.type, variant.value)
        return written

    def write_array(self, array, type_):
        # TODO max array size is 64MiB (67108864 bytes)
        written = self.align(4)
        # length placeholder
        offset = len(self.buffer)
        written += self.write_uint32(0)
        child_type = type_.children[0]

        if child_type.token in 'xtd{(':
            # the first alignment is not included in array size
            written += self.align(8)

        array_len = 0
        if child_type.token == '{':
            for key, value in array.items():
                array_len += self.write_dict_entry([key, value], child_type)
        elif child_type.token == 'y':
            array_len = len(array)
            self.buffer.extend(array)
        else:
            for value in array:
                array_len += self.write_single(child_type, value)

        array_len_packed = pack('<I', array_len)
        for i in range(offset, offset + 4):
            self.buffer[i] = array_len_packed[i - offset]

        return written + array_len

    def write_struct(self, array, type_):
        written = self.align(8)
        for i, value in enumerate(array):
            written += self.write_single(type_.children[i], value)
        return written

    def write_dict_entry(self, dict_entry, type_):
        written = self.align(8)
        written += self.write_single(type_.children[0], dict_entry[0])
        written += self.write_single(type_.children[1], dict_entry[1])
        return written

    def write_single(self, type_, body):
        t = type_.token

        if t not in self.writers:
            raise NotImplementedError(f'type isnt implemented yet: "{t}"')

        return self.writers[t](body, type_)

    def marshall(self):
        self.buffer.clear()
        for i, type_ in enumerate(self.signature_tree.types):
            self.write_single(type_, self.body[i])
        return self.buffer
